/*
  File name: File3Dg.cpp
  Copyright: (C) 2007 Tomasz Kazmierczak

  Creation date: 20.04.2007
  Last modification date: 22.08.2007

  This file is part of Cytadela

 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 3 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *

*/

#include "File3Dg.h"

#define FILEID_3DG 0x44334C4D

//object types stored in file
enum {
	OBJECT3DG_LIGHT,
	OBJECT3DG_TEXTURE,
	OBJECT3DG_MATERIAL,
	OBJECT3DG_MESHBASE,
	OBJECT3DG_MESH
};

bool File3Dg::write(const char *sz3DgFileName)
{
	FILE *p3DgFile = fopen(sz3DgFileName, "wb");
	if(not p3DgFile)
	{
		fprintf(stderr, "ERROR: File3Dg::write() - could not open \'%s\' for writing", sz3DgFileName);
		return false;
	}
	if(not write(p3DgFile))
	{
		fprintf(stderr, "\twhile writing \'%s\'", sz3DgFileName);
		fclose(p3DgFile);
		remove(sz3DgFileName);
		return false;
	}

	fclose(p3DgFile);
	return true;
}

bool File3Dg::write(FILE *p3DgFile)
{
	//file header
	FileHeader3Dg s3DgHeader;
	memset(&s3DgHeader, 0, sizeof(FileHeader3Dg));

	//fill the header
	s3DgHeader.dwId       = FILEID_3DG;  //identifier
	s3DgHeader.dwVersion  = 40;   //version

	//write the header
	if(fwrite(&s3DgHeader, 1, sizeof(FileHeader3Dg), p3DgFile) < sizeof(FileHeader3Dg))
	{
		fprintf(stderr, "ERROR: File3Dg::write() - could not write into the file");
		return false;
	}

	//NOTE: the order of objects IS IMPORTANT

	//lights
	uint32_t uiObjectType = OBJECT3DG_LIGHT;
	for(uint32_t i = 0; i < mLightsQuantity; i++)
	{
		//first write the object type
		fwrite(&uiObjectType, 1, sizeof(uiObjectType), p3DgFile);
		if(fwrite(mLights[i], 1, sizeof(Light3Dg), p3DgFile) < sizeof(Light3Dg))
		{
			fprintf(stderr, "ERROR: File3Dg::write() - could not write into the file");
			return false;
		}
	}
	//textures
	uiObjectType = OBJECT3DG_TEXTURE;
	for(uint32_t i = 0; i < mTexturesQuantity; i++)
	{
		//first write the object type
		fwrite(&uiObjectType, 1, sizeof(uiObjectType), p3DgFile);

		//texture information
		fwrite(&mTextures[i]->eType, 1, sizeof(TextureType3Dg), p3DgFile);
		fwrite(&mTextures[i]->eSWrapMode, 1, sizeof(TextureWrapMode3Dg), p3DgFile);
		fwrite(&mTextures[i]->eTWrapMode, 1, sizeof(TextureWrapMode3Dg), p3DgFile);
		if(fwrite(&mTextures[i]->uiPathBufferSize, 1, sizeof(uint16_t), p3DgFile) < sizeof(uint16_t))
		{
			fprintf(stderr, "ERROR: File3Dg::write() - could not write into the file");
			return false;
		}
		//texture name
		uint16_t uiSize = mTextures[i]->uiPathBufferSize * sizeof(char);
		if(fwrite(mTextures[i]->pPath, 1, uiSize, p3DgFile) < uiSize)
		{
			fprintf(stderr, "ERROR: File3Dg::write() - could not write into the file");
			return false;
		}
	}
	//materials
	uiObjectType = OBJECT3DG_MATERIAL;
	for(uint32_t i = 0; i < mMaterialsQuantity; i++)
	{
		//first write the object type
		fwrite(&uiObjectType, 1, sizeof(uiObjectType), p3DgFile);

		fwrite(&mMaterials[i]->sDiffuse, 1, sizeof(ColorRGBA3Dg), p3DgFile);
		fwrite(&mMaterials[i]->sAmbient, 1, sizeof(ColorRGBA3Dg), p3DgFile);
		fwrite(&mMaterials[i]->sSpecular, 1, sizeof(ColorRGBA3Dg), p3DgFile);
		fwrite(&mMaterials[i]->sEmissive, 1, sizeof(ColorRGBA3Dg), p3DgFile);
		fwrite(&mMaterials[i]->fShininess, 1, sizeof(float), p3DgFile);
		if(fwrite(&mMaterials[i]->uiTexturesQuantity, 1, sizeof(uint32_t), p3DgFile) < sizeof(uint32_t))
		{
			fprintf(stderr, "ERROR: File3Dg::write() - could not write into the file");
			return false;
		}

		//write texture numbers
		for(uint32_t j = 0; j < mMaterials[i]->uiTexturesQuantity; j++)
		{
			uint32_t uiTexNum = mMaterials[i]->getTexture(j);
			if(fwrite(&uiTexNum, 1, sizeof(uint32_t), p3DgFile) < sizeof(uint32_t))
			{
				fprintf(stderr, "ERROR: File3Dg::write() - could not write into the file");
				return false;
			}
		}
	}
	//mesh bases
	uiObjectType = OBJECT3DG_MESHBASE;
	for(MeshBaseMap3Dg::iterator i = mMeshBases.begin(); i != mMeshBases.end(); i++)
	{
		//first write the object type
		fwrite(&uiObjectType, 1, sizeof(uiObjectType), p3DgFile);

		//gather the base's data
		MeshBase3Dg *pMeshBase = i->second;
		MeshBase3DgUID uiMBUID = pMeshBase->getUID();
		uint32_t uiTriansQuantity = pMeshBase->getTriansQuantity();

		//write
		fwrite(&uiMBUID, 1, sizeof(MeshBase3DgUID), p3DgFile);
		if(fwrite(&uiTriansQuantity, 1, sizeof(uint32_t), p3DgFile) < sizeof(uint32_t))
		{
			fprintf(stderr, "ERROR: File3Dg::write() - could not write into the file");
			return false;
		}

		//triangles
		for(uint32_t t = 0; t < uiTriansQuantity; t++)
		{
			Triangle3Dg *pTriangle = pMeshBase->getTriangle(t);
			if(fwrite(pTriangle, 1, sizeof(Triangle3Dg), p3DgFile) < sizeof(Triangle3Dg))
			{
				fprintf(stderr, "ERROR: File3Dg::write() - could not write into the file");
				return false;
			}
		}
	}
	//meshes
	uiObjectType = OBJECT3DG_MESH;
	for(MeshMap3Dg::iterator i = mMeshes.begin(); i != mMeshes.end(); i++)
	{
		//first write the object type
		fwrite(&uiObjectType, 1, sizeof(uiObjectType), p3DgFile);

		//gather the mesh's data
		Mesh3Dg *pMesh = i->second;
		Mesh3DgUID uiMUID = pMesh->getUID();
		MeshBase3DgUID uiMeshBaseUID = 0;
		if(pMesh->getBase())
			uiMeshBaseUID = pMesh->getBase()->getUID();
		uint32_t uiProps = pMesh->getProps();

		//write
		fwrite(&uiMUID, 1, sizeof(Mesh3DgUID), p3DgFile);
		fwrite(&uiMeshBaseUID, 1, sizeof(MeshBase3DgUID), p3DgFile);
		fwrite(&uiProps, 1, sizeof(uint32_t), p3DgFile);
		fwrite(&pMesh->mRotation, 1, sizeof(Matrix3Dg), p3DgFile);
		fwrite(&pMesh->mTranslation, 1, sizeof(Matrix3Dg), p3DgFile);
		fwrite(&pMesh->mScaling, 1, sizeof(Matrix3Dg), p3DgFile);
		fwrite(&pMesh->mMaterialNumber, 1, sizeof(uint32_t), p3DgFile);
		if(fwrite(&pMesh->mAnimation, 1, sizeof(MeshAnim3Dg), p3DgFile) < sizeof(MeshAnim3Dg))
		{
			fprintf(stderr, "ERROR: File3Dg::write() - could not write into the file");
			return false;
		}
	}
	return true;
}

bool File3Dg::read(const char *sz3DgFileName)
{
	FILE *p3DgFile = fopen(sz3DgFileName, "rb");
	if(not p3DgFile)
	{
		fprintf(stderr, "ERROR: File3Dg::read() - could not open \'%s\' for reading", sz3DgFileName);
		return false;
	}
	if(not read(p3DgFile))
	{
		fprintf(stderr, "\twhile reading \'%s\'", sz3DgFileName);
		fclose(p3DgFile);
		return false;
	}
	fclose(p3DgFile);
	return true;
}

bool File3Dg::readLight(FILE *p3DgFile)
{
	Light3Dg *pLight = createLight();
	if(not pLight)
		return false;

	if(fread(pLight, 1, sizeof(Light3Dg), p3DgFile) < sizeof(Light3Dg))
	{
		fprintf(stderr, "ERROR: File3Dg::readLight() - could not read the file");
		return false;
	}
	return true;
}

bool File3Dg::readTexture(FILE *p3DgFile)
{
	//texture info
	TextureType3Dg eTexType;
	TextureWrapMode3Dg eSWrap, eTWrap;
	uint16_t uiPathBufferSize;
	fread(&eTexType, 1, sizeof(TextureType3Dg), p3DgFile);
	fread(&eSWrap, 1, sizeof(TextureWrapMode3Dg), p3DgFile);
	fread(&eTWrap, 1, sizeof(TextureWrapMode3Dg), p3DgFile);
	if(fread(&uiPathBufferSize, 1, sizeof(uint16_t), p3DgFile) < sizeof(uint16_t))
	{
		fprintf(stderr, "ERROR: File3Dg::readTexture() - could not read the file");
		return false;
	}

	char *pPath = new char[uiPathBufferSize];
	if(not pPath)
	{
		fprintf(stderr, "ERROR: File3Dg::readTexture() - out of memory");
		return false;
	}
	//read the texture name
	if(fread(pPath, 1, uiPathBufferSize * sizeof(char), p3DgFile) < uiPathBufferSize * sizeof(char))
	{
		delete pPath;
		fprintf(stderr, "ERROR: File3Dg::readTexture() - could not read the file");
		return false;
	}

	Texture3Dg *pTexture = createTexture(pPath, uiPathBufferSize);
	if(not pTexture)
	{
		delete pPath;
		fprintf(stderr, "ERROR: File3Dg::readTexture() - createTexture() returned \'%s\'", getLastErrorDesc());
		return false;
	}
	delete pPath;

	pTexture->eType = eTexType;
	pTexture->eSWrapMode = eSWrap;
	pTexture->eTWrapMode = eTWrap;
	return true;
}

bool File3Dg::readMaterial(FILE *p3DgFile)
{
	Material3Dg *pMtl = createMaterial();
	if(not pMtl)
	{
		fprintf(stderr, "ERROR: File3Dg::readMaterial() - createMaterial() returned \'%s\'", getLastErrorDesc());
		return false;
	}

	fread(&pMtl->sDiffuse, 1, sizeof(ColorRGBA3Dg), p3DgFile);
	fread(&pMtl->sAmbient, 1, sizeof(ColorRGBA3Dg), p3DgFile);
	fread(&pMtl->sSpecular, 1, sizeof(ColorRGBA3Dg), p3DgFile);
	fread(&pMtl->sEmissive, 1, sizeof(ColorRGBA3Dg), p3DgFile);
	fread(&pMtl->fShininess, 1, sizeof(float), p3DgFile);
	uint32_t uiTexturesQuantity;
	if(fread(&uiTexturesQuantity, 1, sizeof(uint32_t), p3DgFile) < sizeof(uint32_t))
	{
		fprintf(stderr, "ERROR: File3Dg::readMaterial() - could not read the file");
		return false;
	}

	//read texture numbers
	for(uint32_t j = 0; j < uiTexturesQuantity; j++)
	{
		uint32_t uiTexNum;
		if(fread(&uiTexNum, 1, sizeof(uint32_t), p3DgFile) < sizeof(uint32_t))
		{
			fprintf(stderr, "ERROR: File3Dg::readMaterial() - could not read the file");
			return false;
		}
		if(not pMtl->addTexture(uiTexNum))
		{
			fprintf(stderr, "ERROR: File3Dg::readMaterial() - could not read the file");
			return false;
		}
	}
	return true;
}

bool File3Dg::readMeshBase(FILE *p3DgFile)
{
	//read the unique identifier
	MeshBase3DgUID uiUID;
	if(fread(&uiUID, 1, sizeof(MeshBase3DgUID), p3DgFile) < sizeof(MeshBase3DgUID))
	{
		fprintf(stderr, "ERROR: File3Dg::readMeshBase() - could not read the file");
		return false;
	}

	//create a new mesh base
	MeshBase3Dg *pBase = createMeshBase(uiUID);
	if(not pBase)
	{
		fprintf(stderr, "ERROR: File3Dg::readMeshBase() - createMeshBase() returned \'%s\'", getLastErrorDesc());
		return false;
	}

	uint32_t uiTriansQuantity;
	if(fread(&uiTriansQuantity, 1, sizeof(uint32_t), p3DgFile) < sizeof(int32_t))
	{
		fprintf(stderr, "ERROR: File3Dg::readMeshBase() - could not read the file");
		return false;
	}

	//read the trians
	Triangle3Dg sTrian;
	for(uint32_t t = 0; t < uiTriansQuantity; t++)
	{
		if(fread(&sTrian, 1, sizeof(Triangle3Dg), p3DgFile) < sizeof(Triangle3Dg))
		{
			fprintf(stderr, "ERROR: File3Dg::readMeshBase() - could not read the file");
			return false;
		}
		if(not pBase->addTrian(&sTrian))
		{
			fprintf(stderr, "ERROR: File3Dg::readMeshBase() - out of memory");
			return false;
		}
	}
	return true;
}

bool File3Dg::readMesh(FILE *p3DgFile)
{
	//read the unique identifier
	Mesh3DgUID uiUID;
	if(fread(&uiUID, 1, sizeof(Mesh3DgUID), p3DgFile) < sizeof(Mesh3DgUID))
	{
		fprintf(stderr, "ERROR: File3Dg::readMesh() - could not read the file");
		return false;
	}

	//create a new mesh
	Mesh3Dg *pMesh = createMesh(uiUID);
	if(not pMesh)
	{
		fprintf(stderr, "ERROR: File3Dg::readMesh() - createMesh() returned \'%s\'", getLastErrorDesc());
		return false;
	}

	//read the data
	MeshBase3DgUID uiMeshBaseUID;
	uint32_t uiProps;
	fread(&uiMeshBaseUID, 1, sizeof(MeshBase3DgUID), p3DgFile);
	fread(&uiProps, 1, sizeof(uint32_t), p3DgFile);
	fread(&pMesh->mRotation, 1, sizeof(Matrix3Dg), p3DgFile);
	fread(&pMesh->mTranslation, 1, sizeof(Matrix3Dg), p3DgFile);
	fread(&pMesh->mScaling, 1, sizeof(Matrix3Dg), p3DgFile);
	fread(&pMesh->mMaterialNumber, 1, sizeof(int32_t), p3DgFile);
	if(fread(&pMesh->mAnimation, 1, sizeof(MeshAnim3Dg), p3DgFile) < sizeof(MeshAnim3Dg))
	{
		fprintf(stderr, "ERROR: File3Dg::readMesh() - could not read the file");
		return false;
	}

	//info about mesh
	if(uiProps & MRP_HAS_BASE)
		pMesh->setBase(mMeshBases[uiMeshBaseUID]);
	pMesh->setProps(uiProps);
	return true;
}

bool File3Dg::read(FILE *p3DgFile)
{
	//read the header
	FileHeader3Dg s3DgHeader;
	if(fread(&s3DgHeader, 1, sizeof(FileHeader3Dg), p3DgFile) < sizeof(FileHeader3Dg))
	{
		fprintf(stderr, "ERROR: File3Dg::read() - could not read the file");
		return false;
	}

	//check for identifier
	if(s3DgHeader.dwId != FILEID_3DG)
	{
		fprintf(stderr, "ERROR: File3Dg::read() - not a 3Dg file");
		return false;
	}
	//check the version
	if(s3DgHeader.dwVersion != 40)
	{
		fprintf(stderr, "ERROR: File3Dg::read() - bad file version");
		return false;
	}

	//read all objects
	uint32_t uiObjectType;
	while(fread(&uiObjectType, 1, sizeof(uiObjectType), p3DgFile) == sizeof(uiObjectType))
	{
		switch(uiObjectType)
		{
			case OBJECT3DG_LIGHT:
			{
				if(not readLight(p3DgFile))
					return false;
			}
			break;
			case OBJECT3DG_TEXTURE:
			{
				if(not readTexture(p3DgFile))
					return false;
			}
			break;
			case OBJECT3DG_MATERIAL:
			{
				if(not readMaterial(p3DgFile))
					return false;
			}
			break;
			case OBJECT3DG_MESH:
			{
				if(not readMesh(p3DgFile))
					return false;
			}
			break;
			case OBJECT3DG_MESHBASE:
			{
				if(not readMeshBase(p3DgFile))
					return false;
			}
			break;
			default:
				fprintf(stderr, "ERROR: File3Dg::read() - unknown object type");
			return false;
		}
	}
	return true;
}
